]> Shamusworld >> Repos - rmac/blobdiff - sect.h
Partial fix for bug #108 (Fixup cleanups).
[rmac] / sect.h
diff --git a/sect.h b/sect.h
index f2aee8da27ac65d4fe504a08ac4449d9a1a9f859..b3f6e0f6bfdc04acfe87555e6e763e7d4f1dcf4d 100644 (file)
--- a/sect.h
+++ b/sect.h
@@ -1,9 +1,9 @@
 //
 // RMAC - Reboot's Macro Assembler for all Atari computers
 // SECT.H - Code Generation, Fixups and Section Management
-// Copyright (C) 199x Landon Dyer, 2011-2017 Reboot and Friends
+// Copyright (C) 199x Landon Dyer, 2011-2018 Reboot and Friends
 // RMAC derived from MADMAC v1.07 Written by Landon Dyer, 1986
-// Source utilised with the kind permission of Landon Dyer
+// Source utilized with the kind permission of Landon Dyer
 //
 
 #ifndef __SECT_H__
@@ -11,7 +11,7 @@
 
 #include "rmac.h"
 
-// Macros to deposit code in the current section
+// Macros to deposit code in the current section (in Big Endian)
 // D_rword deposits a "6502" format (low, high) word (01).
 #define D_byte(b)      {*chptr++=(uint8_t)(b); sloc++; ch_size++; \
                                                if(orgactive) orgaddr++;}
                                                *chptr++=(uint8_t)((lw)>>8); \
                                                *chptr++=(uint8_t)(lw); \
                                                sloc += 4; ch_size += 4; if(orgactive) orgaddr += 4;}
+#define D_quad(qw)     {*chptr++=(uint8_t)((qw)>>56); \
+                                               *chptr++=(uint8_t)((qw)>>48);\
+                                               *chptr++=(uint8_t)((qw)>>40);\
+                                               *chptr++=(uint8_t)((qw)>>32);\
+                                               *chptr++=(uint8_t)((qw)>>24);\
+                                               *chptr++=(uint8_t)((qw)>>16);\
+                                               *chptr++=(uint8_t)((qw)>>8); \
+                                               *chptr++=(uint8_t)(qw); \
+                                               sloc += 8; ch_size += 8; if(orgactive) orgaddr += 8;}
 #define D_rword(w)     {*chptr++=(uint8_t)(w); *chptr++=(uint8_t)((w)>>8); \
                                                sloc+=2; ch_size+=2;if(orgactive) orgaddr += 2;}
-#define D_single(w) {chcheck(4);*chptr++ = ((char *)&w)[3]; \
-                                               *chptr++ = ((char *)&w)[2]; \
-                                               *chptr++ = ((char *)&w)[1]; \
-                                               *chptr++=((char *)&w)[0]; \
-                                               sloc+=4; ch_size += 4; if(orgactive) orgaddr += 4;}
-#define D_double(w) {chcheck(8);*chptr++=(char)((*(unsigned long long *)&w)); \
-                                               *chptr++=(char)((*(unsigned long long *)&w)>>8); \
-                                               *chptr++=(char)((*(unsigned long long *)&w)>>16); \
-                                               *chptr++=(char)((*(unsigned long long *)&w)>>24); \
-                                               *chptr++=(char)((*(unsigned long long *)&w)>>32); \
-                                               *chptr++=(char)((*(unsigned long long *)&w)>>40); \
-                                               *chptr++=(char)((*(unsigned long long *)&w)>>48); \
-                                               *chptr++=(char)((*(unsigned long long *)&w)>>56); \
-                                               sloc+=8; ch_size += 8; if(orgactive) orgaddr += 8;}
-#ifdef _MSC_VER
-#define D_extend(w) {chcheck(12); *chptr++ = (char)((*(unsigned long long *)&w) >> 56); \
-*chptr++ = (char)(((*(unsigned long long *)&w) >> (52)) & 0xf); \
-*chptr++ = (char)(0); \
-*chptr++ = (char)(0); \
-*chptr++ = (char)(((*(unsigned long long *)&w) >> (48 - 3))|0x80 /* assume that the number is normalised */); \
-*chptr++ = (char)((*(unsigned long long *)&w) >> (40 - 3)); \
-*chptr++ = (char)((*(unsigned long long *)&w) >> (32 - 3)); \
-*chptr++ = (char)((*(unsigned long long *)&w) >> (24 - 3)); \
-*chptr++ = (char)((*(unsigned long long *)&w) >> (16 - 3)); \
-*chptr++ = (char)((*(unsigned long long *)&w) >> (8 - 3)); \
-*chptr++ = (char)((*(unsigned long long *)&w << 3)); \
-*chptr++=(char)(0); \
-       sloc+=12; ch_size += 12; if(orgactive) orgaddr += 12;}
-#elif defined(LITTLE_ENDIAN)
-#define D_extend(w) {chcheck(12);*chptr++=((char *)&w)[0]; \
-                                               *chptr++=((char *)&w)[1]; \
-                                               *chptr++=((char *)&w)[2]; \
-                                               *chptr++=((char *)&w)[3]; \
-                                               *chptr++=((char *)&w)[4]; \
-                                               *chptr++=((char *)&w)[5]; \
-                                               *chptr++=((char *)&w)[6]; \
-                                               *chptr++=((char *)&w)[7]; \
-                                               *chptr++=((char *)&w)[8]; \
-                                               *chptr++=((char *)&w)[9]; \
-                                               *chptr++=((char *)&w)[10]; \
-                                               *chptr++=((char *)&w)[11]; \
-                                               sloc+=12; ch_size += 12; if(orgactive) orgaddr += 12;}
-
-#else
-
-#error Please implement a non-byte swapped D_extend!
-
-#endif
+
+// Macro for the 56001. Word size on this device is 24 bits wide. I hope that
+// orgaddr += 1 means that the addresses in the device reflect this.
+#define D_dsp(w)    {chcheck(3);*chptr++=(uint8_t)(w>>16); \
+       *chptr++=(uint8_t)(w>>8); *chptr++=(uint8_t)w; \
+       sloc+=1; ch_size += 3; if(orgactive) orgaddr += 1; \
+       dsp_written_data_in_current_org=1;}
+
+// This macro expects to get an array of uint8_ts with the hi bits in a[0] and
+// the low bits in a[11] (Big Endian).
+#define D_extend(a) {memcpy(chptr, a, 12); chptr+=12; sloc+=12, ch_size+=12;\
+                                       if (orgactive) orgaddr+=12;}
+
 // Fill n bytes with zeroes
 #define D_ZEROFILL(n)  {memset(chptr, 0, n); chptr+=n; sloc+=n; ch_size+=n;\
                                                if (orgactive) orgaddr+=n;}
@@ -90,6 +64,7 @@
 #define SABS         0x2000            // Section is absolute
 #define SPIC         0x1000            // Section is position-independent code
 
+// N.B.: THIS IS NO LONGER TRUE
 // Fixup record a WORD of these bits, followed by a loc and then a pointer
 // to a symbol or an ENDEXPR-terminated postfix expression.
 //
 // Chunks are used to hold generated code and fixup records
 #define CHUNK  struct _chunk
 CHUNK {
-       CHUNK * chnext;                         // Next, previous chunks in section
-       CHUNK * chprev;
-       uint32_t chloc;                         // Base addr of this chunk
-       uint32_t challoc;                       // # bytes allocated for chunk
-       uint32_t ch_size;                       // # bytes chunk actually uses
-       uint8_t * chptr;                        // Data for this chunk
+       CHUNK *   chnext;       // Next, previous chunks in section
+       CHUNK *   chprev;
+       uint32_t  chloc;        // Base addr of this chunk
+       uint32_t  challoc;      // # bytes allocated for chunk
+       uint32_t  ch_size;      // # bytes chunk actually uses
+       uint8_t * chptr;        // Data for this chunk
+};
+
+// Fixup records can also hold an expression (if any)
+#define FIXUP struct _fixup
+FIXUP {
+       FIXUP *  next;          // Pointer to next FIXUP
+       uint32_t attr;          // Fixup type
+       uint32_t loc;           // Location in section
+       uint16_t fileno;        // ID of current file
+       uint32_t lineno;        // Current line
+       TOKEN *  expr;          // Pointer to stored expression (if any)
+       SYM *    symbol;        // Pointer to symbol (if any)
+       uint32_t orgaddr;       // Fixup origin address (used for FU_JR)
 };
 
 // Section descriptor
 #define SECT   struct _sect
 SECT {
-       uint16_t scattr;                        // Section attributes
-       uint32_t sloc;                          // Current loc-in / size-of section
-       uint32_t relocs;                        // # of relocations for this section
-       CHUNK * sfcode;                         // First chunk in section
-       CHUNK * scode;                          // Last chunk in section
-       CHUNK * sffix;                          // First fixup chunk
-       CHUNK * sfix;                           // Last fixup chunk
+       uint16_t scattr;        // Section attributes
+       uint32_t sloc;          // Current loc-in / size-of section
+       uint32_t relocs;        // # of relocations for this section
+       uint32_t orgaddr;       // Current org'd address ***NEW***
+       CHUNK *  sfcode;        // First chunk in section
+       CHUNK *  scode;         // Last chunk in section
+       FIXUP *  sffix;         // First fixup
+       FIXUP *  sfix;          // Last fixup ***NEW***
 };
 
 // 680x0 defines
@@ -173,6 +162,7 @@ SECT {
 #define FPU_68881 1
 #define FPU_68882 2
 #define FPU_68040 4
+#define FPU_68060 8
 
 // Helper macros to test for active CPU
 #define CHECK00 if (activecpu == CPU_68000) return error(unsupport)
@@ -185,7 +175,7 @@ SECT {
 #define CHECKNO30 if (activecpu != CPU_68030) return error(unsupport)
 #define CHECKNO40 if (activecpu != CPU_68040) return error(unsupport)
 #define CHECKNO60 if (activecpu != CPU_68060) return error(unsupport)
-
+#define CHECKNOFPU if (!activefpu) return error(unsupport)
 
 // Globals, external etc
 extern uint32_t sloc;
@@ -204,7 +194,7 @@ void SwitchSection(int);
 void SaveSection(void);
 int fixtest(int, uint32_t);
 int chcheck(uint32_t);
-int AddFixup(uint16_t, uint32_t, TOKEN *);
+int AddFixup(uint32_t, uint32_t, TOKEN *);
 int ResolveAllFixups(void);
 
 #endif // __SECT_H__